<!DOCTYPE html>
<html lang="en">
    <head>
        <meta charset="UTF-8">
        <link rel="stylesheet" type="text/css" href="{{ style.skin_css }}"  >
        <link rel="stylesheet" type="text/css" href="/static/legacy/style.css">
        <script src="/static/legacy/action_buttons.js"></script>
 
        {% set page_title = 'Reagent Overview' %}
        <title>{{ page_title }}</title>

        {% set bloom_mod = 'lims' %}

    </head>
    <body>   
         {% include 'legacy/bloom_header.html' %}

        <h1>Create New Reagent</h1>
        {% if template_list %}
            <form action="/create_from_template" method="post">
                <label for="template">Choose a Template:</label>
                <select name="euid" id="euid">
                    {% for template in template_list %}<option value="{{ template.euid }}">{{ template.name }}</option>{% endfor %}
                </select>
                <input type="submit" value="Create">
            </form>
        {% else %}
            <p>No templates available for creating new reagent.</p>
        {% endif %}

        <h1>Reagent List</h1>
        {% if instance_list %}
            <ul>
                {% for reagent in instance_list %}
                    <li>
                        <a href="euid_details?euid={{ reagent.euid }}">{{ reagent.name }}</a>
                        <!-- Additional Reagent details can be added here -->
                    </li>
                {% endfor %}
            </ul>
        {% else %}
            <p>No Reagent found.</p>
        {% endif %}
    </body>
</html>
